/* rockywd.c copyright (c) 2000 by Crusader Systems (Pty) Ltd www.crusader.co.za
 * author: Charl P. Botha <cpbotha@ieee.org>
 * $Id: ampro-cmp5e-wd.c,v 1.6 2000/03/15 12:34:38 cpbotha Exp $
 * 
 * thanks go to Crusader Systems (Pty) Ltd who have very graciously allowed 
 * publishing under GPL
 * 
 * linux driver for Ampro CM/P5e PC/104 built-in watchdog (uses SMSC FDC37C67x
 * Enanced Super I/O Controller with Fast IR, and this has watchdog)
 * CM/P5e is at: http://www.ampro.com/products/coremod/cm-p5e.htm
 * This should work with any SMSC FDC37C67x equipped machine, although
 * you might have to tweak things like the CFIDX and CFDTA values.
 * 
 * This watchdog is NOT able to reset the machine independently of hardware.
 * We can configure it to generate an IRQ when timed-out, or an SMI at
 * best.  In this driver, we have an IRQ handler that resets the machine
 * when the watchdog times out.  I don't think Ampro has done very well with 
 * this.  One would expect a good watchdog to be able to reset its machine no
 * matter what the state of the software.  This is sad as the quality of Ampro's
 * PC104 hardware is excellent.
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 * Charl P. Botha <cpbotha@ieee.org> http://cpbotha.net
 * 
 */

#include <linux/config.h>
#include <linux/module.h>
#include <linux/version.h>
#include <linux/types.h>
#include <linux/errno.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/miscdevice.h>
#include <linux/watchdog.h>
#include <linux/malloc.h>
#include <linux/ioport.h>
#include <linux/fcntl.h>
#include <asm/io.h>
#include <asm/uaccess.h>
#include <asm/system.h>
#include <linux/notifier.h>
#include <linux/reboot.h>
#include <linux/init.h>

/* device id */
#define DEVICE_ID 0x40

/* watchdog timeout in seconds */
#define WDT_TIMEOUT 16
/* configuration index port */
#define CFIDX 0x3F0
/* configuration data port */
#define CFDTA 0x3F1
/* we'll set the 36c67x to assert this IRQ at WD timeout; you can change
 * this to anything from 15 down to 1, but not to 2.
 */
#define WDIRQ 5

/* watch dog timer register index definitions */
#define WDT_TIME_OUT_REG 0xF1
#define WDT_VAL_REG 0xF2
#define WDT_CFG_REG 0xF3
#define WDT_CTRL_REG 0xF4

static int cmp5ewd_is_open=0;
/* we need this variable so that if NOWAYOUT is NOT set, and the device
 * is closed, we know not to reset the machine */
static int cmp5ewd_is_closing=0;

/* we try and take over the SMSC IC as fully as possible.  Ampro have been
 * of VERY little help releasing technical information, so we try to rely
 * on their BIOS as little as possible.
 * NOTE: you still need to call cmp5ewd_set_timeout() after this 
 */
static void cmp5ewd_setup_wdt_regs(void) {
   /* start configuration mode */
   outb_p(0x55, CFIDX);

   /* tell the IC we're going to send it logical device number */
   outb_p(0x7, CFIDX);
   /* now select the logical device 8 */
   outb_p(0x08, CFDTA);
   /* set bit 7 on TIME_OUT register, timeout is in seconds */
   outb_p(WDT_TIME_OUT_REG, CFIDX);
   outb_p(0x80, CFDTA);
   
   /* again setup for LD #8 */
   outb_p(0x7, CFIDX);
   outb_p(0x8, CFDTA);
   outb_p(WDT_CFG_REG, CFIDX);
   /* setup to trigger WDIRQ when timeout occurs */
   outb_p(WDIRQ << 4, CFDTA);
   
   /* exit configuration mode */
   outb_p(0xaa, CFIDX);
}

static void cmp5ewd_cleanup_wdt_regs(void) {
   /* start configuration mode */
   outb_p(0x55, CFIDX);
   
   /* select logical device 8 again */
   outb_p(0x7, CFIDX);
   outb_p(0x8, CFDTA);
   /* set WDT_CFG_REG register to 0x0, its default */
   outb_p(WDT_CFG_REG, CFIDX);
   outb_p(0x0, CFDTA);

   /* exit config mode */
   outb_p(0xaa, CFIDX);
}

static void cmp5ewd_set_timeout(int timeout) {
   /* we start configuration mode */
   outb_p(0x55, CFIDX);
   /* idx of LD # config reg, i.e. tell it we're selecting LD# next */
   outb_p(0x7, CFIDX);
   /* logical device 8 */
   outb_p(0x8, CFDTA);
   /* we want to address CRF2 */
   outb_p(WDT_VAL_REG, CFIDX);
   /* set the correct timeout (the ampro BIOS should set us up in secs) */
   outb_p(timeout, CFDTA);
   /* now we have to get out of config mode */
   outb_p(0xaa, CFIDX);
}

/* our lonely little interrupt handler */
void cmp5ewd_interrupt(int irq, void *dev_id, struct pt_regs *regs) {
#ifndef CONFIG_WATCHDOG_NOWAYOUT
   /* if we have a way out, we do this as cleanup_regs causes an interrupt,
    * and that would cause unnecessary reboot with NOWAYOUT NOT set
    */
   if (!cmp5ewd_is_closing)
#endif
#ifndef DEBUG
     machine_restart(NULL);
   printk(KERN_CRIT "D'oh??!  Complain to support@ampro.com about their bad Watchdog.\n");
#else
   printk(KERN_CRIT "The WatchPuppy(tm) would've attempted to reboot now.\n");
#endif
}
 
static int cmp5ewd_open(struct inode *inode, struct file *file) {
   if (MINOR(inode->i_rdev)!=WATCHDOG_MINOR) return -ENODEV;
   /* make sure of exclusive access */
   if (cmp5ewd_is_open) return -EBUSY;
   MOD_INC_USE_COUNT;
   cmp5ewd_is_open=1;
   cmp5ewd_is_closing=0;
   cmp5ewd_setup_wdt_regs();
   /* set timeout to 16 seconds */
   cmp5ewd_setup_wdt_regs();
   cmp5ewd_set_timeout(WDT_TIMEOUT);
   
#ifdef DEBUG
   /* test area begins */
   outb_p(0x55, CFIDX);

   outb_p(0x7, CFIDX);
   outb_p(0x8, CFDTA);
   outb_p(WDT_TIME_OUT_REG, CFIDX);
   printk("WDT_TIME_OUT_REG = 0x%x\n", inb_p(CFDTA));

   outb_p(0x7, CFIDX);   
   outb_p(0x8, CFDTA);
   outb_p(WDT_VAL_REG, CFIDX);
   printk("WDT_VAL_REG = 0x%x\n", inb_p(CFDTA));

   outb_p(0x7, CFIDX);   
   outb_p(0x8, CFDTA);
   outb_p(WDT_CFG_REG, CFIDX);
   printk("WDT_CFG_REG = 0x%x\n", inb_p(CFDTA));
   
   outb_p(0x7, CFIDX);   
   outb_p(0x8, CFDTA);
   outb_p(WDT_CTRL_REG, CFIDX);
   printk("WDT_CTRL_REG = 0x%x\n", inb_p(CFDTA));
   
   outb_p(0xaa, CFIDX);
   /* test area ends */
#endif
   
   return 0;
}

static int cmp5ewd_release(struct inode *inode, struct file *file) {
   if (MINOR(inode->i_rdev)==WATCHDOG_MINOR) {
      cmp5ewd_is_open=0;
      cmp5ewd_is_closing=1;
#ifndef CONFIG_WATCHDOG_NOWAYOUT
      /* disable watchdog by clearing IRQ map and setting timeout to 0 */
      cmp5ewd_cleanup_wdt_regs();
      cmp5ewd_set_timeout(0);
#endif
   }
   MOD_DEC_USE_COUNT;
   return 0;
}

static ssize_t cmp5ewd_write(struct file *file, const char *buf, size_t count, loff_t *ppos) {
   if (ppos!=&file->f_pos)
     return -ESPIPE;
   if (count) {
      /* reset timeout to default so we don't go boom, but we count down
       * once again 
       */
      cmp5ewd_set_timeout(WDT_TIMEOUT);
      return 1;
   }
   return 0;
}

static int cmp5ewd_notify_sys(struct notifier_block *this, unsigned long code,
	void *unused) {
	if(code==SYS_DOWN || code==SYS_HALT) {
	   /* we're being restarted, so let's be kind an deactivate the wd */
	   cmp5ewd_cleanup_wdt_regs();
	   cmp5ewd_set_timeout(0);
	}
	return NOTIFY_DONE;
}
 

static struct file_operations cmp5ewd_fops = {
   NULL,
     NULL,
     cmp5ewd_write,
     NULL,
     NULL,
     NULL,
     NULL,
     cmp5ewd_open,
     NULL,
     cmp5ewd_release
};
     
static struct miscdevice cmp5ewd_miscdev = {
   WATCHDOG_MINOR,
     "watchdog",
     &cmp5ewd_fops
};

static struct notifier_block cmp5ewd_notifier = {
   cmp5ewd_notify_sys,
     NULL,
     0
};

#ifdef MODULE
#define cmp5ewd_init init_module

void cleanup_module(void) {
   /* unregister our driver */
   misc_deregister(&cmp5ewd_miscdev);
   /* don't notify us of reboot */
   unregister_reboot_notifier(&cmp5ewd_notifier);
   /* release resources that we borrowed */
   release_region(CFIDX, 1);
   release_region(CFDTA, 2);
   free_irq(WDIRQ, NULL);
}

#endif

__initfunc(int cmp5ewd_init(void)) {
   unsigned char devid, cfgb0, cfgb1;
   printk(KERN_INFO "Ampro CM/P5e watchdog driver $Revision: 1.6 $: <cpbotha@ieee.org>\n");
   
   /* make sure we're working with the right device; we have to poke a bit */
   /* start config mode */
   outb_p(0x55, CFIDX);
   /* get device ID */
   outb_p(0x20, CFIDX);
   devid = inb_p(CFDTA);
   /* get config port */
   outb_p(0x26, CFIDX);
   cfgb0 = inb_p(CFDTA);
   outb_p(0x27, CFIDX);
   cfgb1 = inb_p(CFDTA);
   /* end config mode */
   outb_p(0xaa, CFIDX);
   /* compare to expected values */
   if (devid!=DEVICE_ID || ((cfgb1 << 8) | cfgb0)!=CFIDX) {
      printk("Either you don't have an SMSC FDC37C67x, or it's hiding.");
      return -EIO;
   }
   
   /* ask kernel if we can use this irq */
   if (request_irq(WDIRQ, cmp5ewd_interrupt, SA_INTERRUPT, "cmp5ewd", NULL)) {
      printk("Could not reserve IRQ %d!\n", WDIRQ);
      return -EIO;
   }
   /* register ourselves and file-operations struct */
   misc_register(&cmp5ewd_miscdev);
   /* take some memory regions */
   request_region(CFIDX, 1, "cmp5ewd");
   request_region(CFDTA, 2, "cmp5ewd");
   /* make sure we get told about reboots and things */
   register_reboot_notifier(&cmp5ewd_notifier);
   
   return 0;
}
